//-----------------------------------------------------------------------------
// Copyright (c) 2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Dhpoware;
using System.Collections.Generic;

namespace XNABillboard4
{
    /// <summary>
    /// This XNA Windows application builds on the XNA 4.0 Billboarding 3 Demo
    /// (http://www.dhpoware.com/demos/xnaBillboards3.html) by showing you how
    /// to render lots of billboards. The naive single pass rendering technique
    /// used in the XNA 4.0 Billboarding 3 Demo won't work correctly in this
    /// demo. This is because the depth values of each billboard's transparent
    /// pixels are written to the depth buffer causing the billboards to be
    /// incorrectly rendered. The correct solution is to sort the billboards
    /// in a back-to-front order prior to rendering. However doing so is slow.
    /// A commonly used alternative that produces mostly correct results and
    /// is much faster is to use a two pass rendering technique. In the first
    /// pass render the non-transparent pixels of the billboards with blending
    /// disabled and alpha-testing enabled. In the second pass render the
    /// transparent pixels of the billboards with blending enabled and depth
    /// buffer writes disabled. Further details on this rendering technique
    /// can be found in this Microsoft XNA Billboard sample:
    /// http://create.msdn.com/en-US/education/catalog/sample/billboard.
    /// </summary>
    public class Demo : Microsoft.Xna.Framework.Game
    {
        private static void Main(string[] args)
        {
            using (Demo game = new Demo())
            {
                game.Run();
            }
        }

        private enum BillboardingTechnique
        {
            CameraAligned,
            WorldYAxisAligned
        }

        private static readonly float BILLBOARD_ANIM_FPS = 24.0f;
        private static readonly float BILLBOARD_ANIM_TIME_MIN = 0.0f;
        private static readonly float BILLBOARD_ANIM_TIME_MAX = MathHelper.TwoPi;
        private static readonly float BILLBOARD_ANIM_SCALE = 0.15f;

        private CameraComponent cameraComponent;
        private GridComponent gridComponent;
        private GraphicsDeviceManager graphics;
        private SpriteBatch spriteBatch;
        private SpriteFont spriteFont;
        private KeyboardState currentKeyboardState;
        private KeyboardState prevKeyboardState;
        private int windowWidth;
        private int windowHeight;
        private int frames;
        private int framesPerSecond;
        private bool displayHelp;
        private TimeSpan elapsedTime = TimeSpan.Zero;
        private Vector2 billboardSize = new Vector2(2.0f, 2.0f);
        private Billboard billboard;
        private Texture2D billboardTexture;
        private Effect billboardEffect;
        private BillboardingTechnique billboardingTechnique;
        private TimeSpan billboardAnimationElapsedTime = TimeSpan.Zero;
        private float billboardAnimationTime = BILLBOARD_ANIM_TIME_MIN;

        public Demo()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            cameraComponent = new CameraComponent(this);
            Components.Add(cameraComponent);

            gridComponent = new GridComponent(this);
            Components.Add(gridComponent);

            Window.Title = "XNA 4.0 Billboards 4 Demo";
            IsFixedTimeStep = false;
        }

        protected override void Initialize()
        {
            base.Initialize();

            // Setup the window to be a quarter the size of the desktop.
            windowWidth = GraphicsDevice.DisplayMode.Width / 2;
            windowHeight = GraphicsDevice.DisplayMode.Height / 2;

            // Setup frame buffer.
            graphics.SynchronizeWithVerticalRetrace = false;
            graphics.PreferredBackBufferWidth = windowWidth;
            graphics.PreferredBackBufferHeight = windowHeight;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();

            // Setup the initial input states.
            currentKeyboardState = Keyboard.GetState();

            // Setup the GridComponent.
            gridComponent.CreateGrid(Color.LightGray, Vector3.Zero, 16, 1);

            // Setup the camera.
            cameraComponent.CurrentBehavior = Camera.Behavior.Orbit;
            cameraComponent.ClickAndDragMouseRotation = true;
            cameraComponent.Perspective(90.0f, (float)windowWidth / (float)windowHeight, 0.1f, 3.0f * gridComponent.GridSize);
            cameraComponent.LookAt(new Vector3(0.0f, gridComponent.GridSize * 0.25f, gridComponent.GridSize * 0.5f), Vector3.Zero, Vector3.Up);
            cameraComponent.OrbitMinZoom = gridComponent.GridSize * 0.25f;
            cameraComponent.OrbitMaxZoom = gridComponent.GridSize * 1.5f;
            cameraComponent.OrbitOffsetDistance = cameraComponent.Position.Length();

            // Setup the billboard.

            List<Vector3> positions = new List<Vector3>();
            float gridSize = gridComponent.GridSize * gridComponent.GridSpacing;
            bool fullRow = true;

            for (float z = -gridSize * 0.5f; z <= gridSize * 0.5f;  z += billboardSize.X * 0.25f)
            {
                float offset = !fullRow ? billboardSize.X * 0.5f : 0.0f;

                for (float x = -gridSize * 0.5f + offset; x <= gridSize * 0.5f - offset; x += billboardSize.X)
                    positions.Add(new Vector3(x, billboardSize.Y * 0.5f, z));

                fullRow = !fullRow;
            }

            billboard = new Billboard(GraphicsDevice,
                                      positions.ToArray(),
                                      1.0f,
                                      1.0f,
                                      0.0f,
                                      0.0f);
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
            
            // Load the sprite font. The sprite font has a 3 pixel outer glow
            // baked into it so we need to decrease the spacing so that the
            // SpriteFont will render correctly.
            spriteFont = Content.Load<SpriteFont>(@"Fonts\DemoFont14pt");
            spriteFont.Spacing = -4.0f;

            // Load texture for the billboard.
            billboardTexture = Content.Load<Texture2D>(@"Textures\grass");

            // Load the Effect file that implements the billboarding shaders.
            billboardEffect = Content.Load<Effect>(@"Effects\Billboard");
            billboardEffect.CurrentTechnique = billboardEffect.Techniques["BillboardingWorldYAxisAligned"];
            billboardingTechnique = BillboardingTechnique.WorldYAxisAligned;
        }

        protected override void UnloadContent()
        {
        }

        private void ToggleFullScreen()
        {
            int newWidth = 0;
            int newHeight = 0;

            graphics.IsFullScreen = !graphics.IsFullScreen;

            if (graphics.IsFullScreen)
            {
                newWidth = GraphicsDevice.DisplayMode.Width;
                newHeight = GraphicsDevice.DisplayMode.Height;
            }
            else
            {
                newWidth = windowWidth;
                newHeight = windowHeight;
            }

            graphics.PreferredBackBufferWidth = newWidth;
            graphics.PreferredBackBufferHeight = newHeight;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();
        }

        private bool KeyJustPressed(Keys key)
        {
            return currentKeyboardState.IsKeyDown(key) && prevKeyboardState.IsKeyUp(key);
        }

        private void ProcessKeyboard()
        {
            prevKeyboardState = currentKeyboardState;
            currentKeyboardState = Keyboard.GetState();

            if (KeyJustPressed(Keys.Escape))
                this.Exit();

            if (currentKeyboardState.IsKeyDown(Keys.LeftAlt) || currentKeyboardState.IsKeyDown(Keys.RightAlt))
            {
                if (KeyJustPressed(Keys.Enter))
                    ToggleFullScreen();
            }

            if (KeyJustPressed(Keys.H))
                displayHelp = !displayHelp;

            if (KeyJustPressed(Keys.D1))
            {
                billboardingTechnique = BillboardingTechnique.CameraAligned;
                billboardEffect.CurrentTechnique = billboardEffect.Techniques["BillboardingCameraAligned"];
            }

            if (KeyJustPressed(Keys.D2))
            {
                billboardingTechnique = BillboardingTechnique.WorldYAxisAligned;
                billboardEffect.CurrentTechnique = billboardEffect.Techniques["BillboardingWorldYAxisAligned"];
            }
        }

        private void UpdateBillboardAnimationTime(GameTime gameTime)
        {
            float frameTime = 1 / BILLBOARD_ANIM_FPS;
            float frameStep = (BILLBOARD_ANIM_TIME_MAX - BILLBOARD_ANIM_TIME_MIN) / BILLBOARD_ANIM_FPS;

            billboardAnimationElapsedTime += gameTime.ElapsedGameTime;

            if (billboardAnimationElapsedTime > TimeSpan.FromSeconds(frameTime))
            {
                billboardAnimationElapsedTime -= TimeSpan.FromSeconds(frameTime);
                billboardAnimationTime += frameStep;

                if (billboardAnimationTime > BILLBOARD_ANIM_TIME_MAX)
                    billboardAnimationTime = BILLBOARD_ANIM_TIME_MIN;
            }
        }

        private void UpdateFrameRate(GameTime gameTime)
        {
            elapsedTime += gameTime.ElapsedGameTime;

            if (elapsedTime > TimeSpan.FromSeconds(1))
            {
                elapsedTime -= TimeSpan.FromSeconds(1);
                framesPerSecond = frames;
                frames = 0;
            }
        }

        private void IncrementFrameCounter()
        {
            ++frames;
        }

        protected override void Update(GameTime gameTime)
        {
            if (!this.IsActive)
                return;

            base.Update(gameTime);

            ProcessKeyboard();
            UpdateFrameRate(gameTime);
            UpdateBillboardAnimationTime(gameTime);

            gridComponent.ViewMatrix = cameraComponent.ViewMatrix;
            gridComponent.ProjectionMatrix = cameraComponent.ProjectionMatrix;
        }

        private void DrawText()
        {
            string text = null;

            if (displayHelp)
            {
                System.Text.StringBuilder buffer = new System.Text.StringBuilder();

                buffer.AppendLine("Press 1 for camera-aligned billboarding");
                buffer.AppendLine("Press 2 for world y-axis aligned billboarding");
                buffer.AppendLine("Press ALT + ENTER to toggle full screen");
                buffer.AppendLine("Press H to hide help");

                text = buffer.ToString();
            }
            else
            {
                System.Text.StringBuilder buffer = new System.Text.StringBuilder();

                buffer.AppendFormat("FPS: {0}\n", framesPerSecond);
                buffer.AppendFormat("Camera: x:{0} y:{1} z:{2}\n",
                    cameraComponent.Position.X.ToString("#0.00"),
                    cameraComponent.Position.Y.ToString("#0.00"),
                    cameraComponent.Position.Z.ToString("#0.00"));

                switch (billboardingTechnique)
                {
                case BillboardingTechnique.CameraAligned:
                    buffer.AppendLine("Technique: Camera Aligned");
                    break;

                case BillboardingTechnique.WorldYAxisAligned:
                    buffer.AppendLine("Technique: World Y-Axis Aligned");
                    break;

                default:
                    break;
                }

                buffer.AppendLine();
                buffer.Append("Press H to display help");

                text = buffer.ToString();
            }

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
            spriteBatch.DrawString(spriteFont, text, new Vector2(1.0f, 1.0f), Color.Yellow);
            spriteBatch.End();
        }

        /// <summary>
        /// Render the billboards using alpha blending. The naive approach of
        /// just enabling alpha blending and then rendering all billboards in
        /// one pass will not work. That's because when each billboard is drawn
        /// the GPU will update the depth buffer with the depth of each
        /// transparent pixel in the billboard. Thereby giving the incorrect
        /// result. The correct way to render all the billboards requires that
        /// the billboards be rendered in a back to front order where those
        /// billboards furthest from the camera are rendered before those
        /// billboards that are closer to the camera. This basically means
        /// sorting the billboards according to their Z depths every time the
        /// camera changes position and orientation. We don't do that here
        /// because it is too slow. Instead we use a 2 pass technique here that
        /// is much faster and gives mostly correct results. In the first pass
        /// we render the non-transparent pixels of the billboards. We disable
        /// alpha blending and enable alpha testing and depth writes to ensure
        /// that the depths of the non-transparent pixels in the billboards are
        /// written to the depth buffer for use in the next pass. In the second
        /// pass we render the transparent pixels of the billboard. We enable
        /// alpha blending and disable depth writes. The GPU will use the depth
        /// values of the non-transparent pixels written in the previous pass
        /// to ensure the billboards are correctly rendered.
        /// </summary>
        /// <param name="gameTime"></param>
        private void DrawBillboard(GameTime gameTime)
        {
            // Save current states.
            
            RasterizerState prevRasterizerState = GraphicsDevice.RasterizerState;
            BlendState prevBlendState = GraphicsDevice.BlendState;

            // First pass:
            // Render the non-transparent pixels of the billboards and store
            // their depths in the depth buffer.

            billboardEffect.Parameters["world"].SetValue(Matrix.Identity);
            billboardEffect.Parameters["view"].SetValue(cameraComponent.ViewMatrix);
            billboardEffect.Parameters["projection"].SetValue(cameraComponent.ProjectionMatrix);
            billboardEffect.Parameters["billboardSize"].SetValue(billboardSize);
            billboardEffect.Parameters["colorMap"].SetValue(billboardTexture);
            billboardEffect.Parameters["animationTime"].SetValue(billboardAnimationTime);
            billboardEffect.Parameters["animationScaleFactor"].SetValue(BILLBOARD_ANIM_SCALE);
            billboardEffect.Parameters["alphaTestDirection"].SetValue(1.0f);
            
            GraphicsDevice.BlendState = BlendState.Opaque;
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
            GraphicsDevice.RasterizerState = RasterizerState.CullNone;

            billboard.Draw(GraphicsDevice, billboardEffect);

            // Second pass:
            // Render the transparent pixels of the billboards.
            // Disable depth buffer writes to ensure that the depth values from
            // the first pass are used instead.

            billboardEffect.Parameters["alphaTestDirection"].SetValue(-1.0f);
            
            GraphicsDevice.BlendState = BlendState.NonPremultiplied;
            GraphicsDevice.DepthStencilState = DepthStencilState.DepthRead;

            billboard.Draw(GraphicsDevice, billboardEffect);

            // Restore original states.

            GraphicsDevice.BlendState = prevBlendState;
            GraphicsDevice.RasterizerState = prevRasterizerState;
        }

        protected override void Draw(GameTime gameTime)
        {
            if (!this.IsActive)
                return;

            GraphicsDevice.Clear(Color.CornflowerBlue);
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
            
            base.Draw(gameTime);

            DrawBillboard(gameTime);
            DrawText();

            IncrementFrameCounter();
        }
    }
}